#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

__copyright__: str = "Copyright 2021, Yahoo"

__import__("pkg_resources").declare_namespace(__name__)
